/*
** ADM-XRC Series Driver
**
** (C) Copright 2001-2005 Alpha Data
**
** methods_flash.c - Model-specific routines for accessing Flash memory.
*/

#include "drvparam.h"
#include "flash.h"
#include "methods.h"

MethodResult
methodFlashPlxbased(
  DeviceContext* pDevCtx,
  os_boolean     write,
  os_uint32      location,
  os_uint16*     value)
{
  /* Offset (in bytes) of Flash area in control space */
  const uint32_t flashOffset = 0x200000U;
  /* Flash area in control space is 2 MiB in size */
  const uint32_t flashPageMask = 0x1fffffU;
  const unsigned int flashPageShift = 21U;
  CONTROLXRC* pRegs = pDevCtx->hw.control.plxbased;
  OS_BAR_HANDLE handle = pDevCtx->hw.controlHandle;
  os_uint8 page = (os_uint8)(location >> flashPageShift);
  os_uint8* pFlash = (os_uint8*)pRegs + flashOffset;

  os_dbgprint(9, ("methodFlashPlxbased: write=%ld location=%lx *value=%lx\n",
		  (unsigned long)write, (unsigned long)location, (unsigned long)*value));
  
  switch (pDevCtx->devId.pci.subDevice) {
  case PCI_SUBDEVICE_ADATA_ADMXRC2:
  case PCI_SUBDEVICE_ADATA_ADPWRC2:
  case PCI_SUBDEVICE_ADATA_ADPDRC2:
    /* Set Flash page register on ADM-XRCII/ADP-WRCII/ADP-DRCII only */
    os_write_register_le8(handle, &pRegs->variant.admxrc2.FLPAGE, page & 0x7fU);
    os_read_register_le8(handle, &pRegs->variant.admxrc2.FLPAGE);
    break;

  default:
    if (page) {
      return MethodResult_InvalidLocation;
    }
    break;
  }
  
  location &= flashPageMask;
  if (write) {
    os_write_register_le8(handle, pFlash + location, (os_uint8)*value);
    os_read_register_le8(handle, &pRegs->variant.admxrc2.FLPAGE);
  } else {
    *value = (os_uint16)os_read_register_le8(handle, pFlash + location);
  }
  
  return MethodResult_Success;
}

MethodResult
methodFlashXrc2l(
  DeviceContext* pDevCtx,
  os_boolean     write,
  os_uint32      location,
  os_uint16*     value)
{
  return MethodResult_NotPresent;
}

MethodResult
methodFlashXpl(
  DeviceContext* pDevCtx,
  os_boolean     write,
  os_uint32      location,
  os_uint16*     value)
{
  XPLSPACE1REGS* pRegs = pDevCtx->hw.control.xpl;
  OS_BAR_HANDLE handle = pDevCtx->hw.controlHandle;
  os_uint32      flctl, fldata;
  
  os_dbgprint(9, ("methodFlashXpl: write=%ld location=%lx *value=%lx\n",
		  (unsigned long) write, (unsigned long) location, (unsigned long) *value));
  
  if (write) {
    /* Drive fl_a, fl_sel_n, fl_oe_n, fl_we_n, assert fl_sel_n */
    flctl = (location & 0x01ffffffU) | (0x1fU << 25);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    /* Drive fl_d (data to Flash) */
    fldata = (*value & 0xffU) | 0x100U;
    os_write_register_le32(handle, &pRegs->FLDATA, fldata);
    os_read_register_le32(handle, &pRegs->FLCTL);
    /* Assert fl_we_n (write enable for Flash) */
    flctl |= (0x1U << 31);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    os_read_register_le32(handle, &pRegs->FLCTL);
    /* Deassert fl_we_n */
    flctl &= ~(0x1U << 31);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    os_read_register_le32(handle, &pRegs->FLCTL);
    /* Stop driving fl_d (data to Flash) */
    fldata &= ~0x100U;
    os_write_register_le32(handle, &pRegs->FLDATA, fldata);
    /* Stop driving fl_a, fl_sel_n, fl_oe_n, fl_we_n, assert fl_sel_n */
    flctl &= ~(0x1fU << 25);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    os_read_register_le32(handle, &pRegs->FLCTL);
  } else {
    /* Drive fl_a, fl_sel_n, fl_oe_n, fl_we_n, assert fl_sel_n */
    flctl = (location & 0x01ffffffU) | (0x1fU << 25);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    os_read_register_le32(handle, &pRegs->FLCTL);
    /* Assert fl_oe_n (output enable for Flash) */
    flctl |= (0x1U << 30);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    os_read_register_le32(handle, &pRegs->FLCTL);
    /* Read fl_d (data from Flash) */
    *value = (os_uint8) os_read_register_le32(handle, &pRegs->FLDATA);
    /* Stop driving fl_a, fl_sel_n, fl_oe_n, fl_we_n, assert fl_sel_n and deassert fl_oe_n */
    flctl &= ~(0x3fU << 25);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    os_read_register_le32(handle, &pRegs->FLCTL);
    os_dbgprint(9, ("methodFlashXpl: *value=%lx\n", (unsigned long)*value));
  
  }
  
  return MethodResult_Success;
}

MethodResult
methodFlashXrc2n(
  DeviceContext* pDevCtx,
  os_boolean     write,
  os_uint32      location,
  os_uint16*     value)
{
  CONTROLWRC2* pRegs = pDevCtx->hw.control.wrc2;
  OS_BAR_HANDLE handle = pDevCtx->hw.controlHandle;

  os_dbgprint(9, ("methodFlashXrc2n: write=%ld location=%lx *value=%lx\n",
		  (unsigned long) write, (unsigned long) location, (unsigned long) *value));
  
  /* Set up the address */
  os_write_register_le32(handle, &pRegs->FLADDR, location);
  os_read_register_le32(handle, &pRegs->FLADDR);
  
  if (write) {
    /* Write the value */
    os_write_register_le8(handle, &pRegs->FLDATA, (os_uint8) *value);
  } else {
    /* Read the value */
    *value = (os_uint16) os_read_register_le8(handle, &pRegs->FLDATA);
  }
  
  return MethodResult_Success;
}

MethodResult
methodFlashXprev0(
  DeviceContext* pDevCtx,
  os_boolean     write,
  os_uint32      location,
  os_uint16*     value)
{
  return MethodResult_NotPresent;
}

MethodResult
methodFlashXprev1(
  DeviceContext* pDevCtx,
  os_boolean     write,
  os_uint32      location,
  os_uint16*     value)
{
  XPSPACE1REGS* pRegs = pDevCtx->hw.control.xp;
  OS_BAR_HANDLE handle = pDevCtx->hw.controlHandle;
  os_uint32 flctl, fldata;
  os_uint32 val32;
  
  os_dbgprint(9, ("methodFlashXprev1: write=%ld location=%lx *value=%lx\n",
		  (unsigned long) write, (unsigned long) location, (unsigned long) *value));
  
  fldata = pDevCtx->shadow.regs.xp.fldata;
  
  /* Drive fl_a, fl_sel_n, fl_oe_n, fl_we_n, assert fl_sel_n */
  flctl = ((location & 0x1fffffeU) >> 1) | (0x1fU << 25);
  os_write_register_le32(handle, &pRegs->FLCTL, flctl);
  val32 = os_read_register_le32(handle, &pRegs->FLCTL);
  
  if (write) {
    /* Drive fl_d */
    fldata |= (*value | 0x10000U);
    os_write_register_le32(handle, &pRegs->FLDATA, fldata);
    val32 = os_read_register_le32(handle, &pRegs->FLDATA);
    
    /* Assert fl_we_n (write enable for Flash) */
    flctl |= (0x1U << 31);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Deassert fl_we_n */
    flctl &= ~(0x1U << 31);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Deassert fl_sel_n */
    flctl &= ~(0x1U << 29);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
  } else {
    /* Assert fl_oe_n */
    flctl |= (0x1U << 30);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    
    /* Read fl_d */
    *value = (os_uint16) os_read_register_le32(handle, &pRegs->FLDATA);
    
    /* Deassert fl_oe_n */
    flctl &= ~(0x1U << 30);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Deassert fl_sel_n */
    flctl &= ~(0x1U << 29);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    os_dbgprint(9, ("methodFlashXprev1: *value=0x%lx\n",
		    (unsigned long) *value));
  }
  
  /* Stop driving fl_a, fl_sel_n, fl_oe_n, fl_we_n */
  os_write_register_le32(handle, &pRegs->FLCTL, 0);
  os_read_register_le32(handle, &pRegs->FLCTL);
  
  /* Stop driving fl_d */
  fldata &= ~0x10000U;
  os_write_register_le32(handle, &pRegs->FLDATA, fldata);
  os_read_register_le32(handle, &pRegs->FLDATA);
  
  return MethodResult_Success;
}

MethodResult
methodFlashXrc4ls(
  DeviceContext* pDevCtx,
  os_boolean     write,
  os_uint32      location,
  os_uint16*     value)
{
  CONTROLXRC4LS* pRegs = pDevCtx->hw.control.xrc4ls;
  OS_BAR_HANDLE handle = pDevCtx->hw.controlHandle;
  os_uint32 flctl, flpage;
  os_uint32 val32;

  os_dbgprint(9, ("methodFlashXrc4ls: write=%ld location=%lx *value=%lx\n",
		  (unsigned long) write, (unsigned long) location, (unsigned long) *value));

  flpage = location & 0x1fe0000U;
  os_write_register_le32(handle, &pRegs->FLPAGE, flpage);
  val32 = os_read_register_le32(handle, &pRegs->FLPAGE);

  if (write) {
    /* Select mode 0, set fl_a, Drive fl_a, fl_d, fl_ce_n, fl_oe_n, fl_we_n, assert fl_ce_n */
    flctl = (location & 0x1fffeU) | (0x7U << 24) | (0x1U << 28);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Set fl_d */
    os_write_register_le32(handle, &pRegs->FLDATA, *value);
    val32 = os_read_register_le32(handle, &pRegs->FLDATA);
    
    /* Assert fl_we_n */
    flctl |= (0x1U << 30);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Deassert fl_we_n */
    flctl &= ~(0x1U << 30);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Deassert fl_ce_n */
    flctl &= ~(0x1U << 28);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Stop driving fl_a, fl_d, fl_ce_n, fl_oe_n, fl_we_n */
    flctl &= ~(0x7U << 24);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
  } else {
    /* Select mode 0, set fl_a, Drive fl_a, fl_ce_n, fl_oe_n, fl_we_n, assert fl_ce_n */
    flctl = (location & 0x1fffeU) | (0x3U << 25) | (0x1U << 28);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Assert fl_oe_n */
    flctl |= (0x1U << 29);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Read fl_d */
    *value = (os_uint16) os_read_register_le32(handle, &pRegs->FLDATA);
    
    /* Deassert fl_oe_n */
    flctl &= ~(0x1U << 29);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Deassert fl_ce_n */
    flctl &= ~(0x1U << 28);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Stop driving fl_a, fl_ce_n, fl_oe_n, fl_we_n */
    flctl &= ~(0x3U << 25);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    os_dbgprint(9, ("methodFlashXrc4ls: *value=%lx\n",
		    (unsigned long) *value));
  }

  return MethodResult_Success;
}

MethodResult
methodFlashXrc4fx(
  DeviceContext* pDevCtx,
  os_boolean     write,
  os_uint32      location,
  os_uint16*     value)
{
  XRC4FXSPACE1REGS* pRegs = pDevCtx->hw.control.xrc4fx;
  OS_BAR_HANDLE handle = pDevCtx->hw.controlHandle;
  os_uint32 flctl, flpage;
  os_uint32 val32;
  
  os_dbgprint(9, ("methodFlashXrc4fx: write=%ld location=%lx *value=%lx\n",
		  (unsigned long) write, (unsigned long) location, (unsigned long) *value));
  
  flpage = location & 0x3fe0000U;
  os_write_register_le32(handle, &pRegs->FLPAGE, flpage);
  val32 = os_read_register_le32(handle, &pRegs->FLPAGE);
  
  if (write) {
    /* Select mode 0, set fl_a, Drive fl_a, fl_d, fl_ce_n, fl_oe_n, fl_we_n, assert fl_ce_n */
    flctl = (location & 0x1fffeU) | (0x7U << 24) | (0x1U << 28);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Set fl_d */
    os_write_register_le32(handle, &pRegs->FLDATA, *value);
    val32 = os_read_register_le32(handle, &pRegs->FLDATA);
    
    /* Assert fl_we_n */
    flctl |= (0x1U << 30);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Deassert fl_we_n */
    flctl &= ~(0x1U << 30);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Deassert fl_ce_n */
    flctl &= ~(0x1U << 28);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Stop driving fl_a, fl_d, fl_ce_n, fl_oe_n, fl_we_n */
    flctl &= ~(0x7U << 24);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
  } else {
    /* Select mode 0, set fl_a, Drive fl_a, fl_ce_n, fl_oe_n, fl_we_n, assert fl_ce_n */
    flctl = (location & 0x1fffeU) | (0x3U << 25) | (0x1U << 28);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Assert fl_oe_n */
    flctl |= (0x1U << 29);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Read fl_d */
    *value = (os_uint16) os_read_register_le32(handle, &pRegs->FLDATA);
    
    /* Deassert fl_oe_n */
    flctl &= ~(0x1U << 29);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Deassert fl_ce_n */
    flctl &= ~(0x1U << 28);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Stop driving fl_a, fl_ce_n, fl_oe_n, fl_we_n */
    flctl &= ~(0x3U << 25);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    os_dbgprint(9, ("methodFlashXrc4fx: *value=%lx\n",
		    (unsigned long) *value));
  }
  
  return MethodResult_Success;
}

MethodResult
methodFlashXrce4fx(
  DeviceContext* pDevCtx,
  os_boolean     write,
  os_uint32      location,
  os_uint16*     value)
{
  XRCE4FXSPACE1REGS* pRegs = pDevCtx->hw.control.xrce4fx;
  OS_BAR_HANDLE handle = pDevCtx->hw.controlHandle;
  os_uint32 flctl, flpage;
  os_uint32 val32;
  
  os_dbgprint(9, ("methodFlashXrce4fx: write=%ld location=%lx *value=%lx\n",
		  (unsigned long) write, (unsigned long) location, (unsigned long) *value));
  
  flpage = location & 0x1fe0000U;
  os_write_register_le32(handle, &pRegs->FLPAGE, flpage);
  val32 = os_read_register_le32(handle, &pRegs->FLPAGE);
  
  if (write) {
    /* Select mode 0, set fl_a, Drive fl_a, fl_d, fl_ce_n, fl_oe_n, fl_we_n, assert fl_ce_n */
    flctl = (location & 0x1fffeU) | (0x7U << 24) | (0x1U << 28);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Set fl_d */
    os_write_register_le32(handle, &pRegs->FLDATA, *value);
    val32 = os_read_register_le32(handle, &pRegs->FLDATA);
    
    /* Assert fl_we_n */
    flctl |= (0x1U << 30);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Deassert fl_we_n */
    flctl &= ~(0x1U << 30);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Deassert fl_ce_n */
    flctl &= ~(0x1U << 28);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Stop driving fl_a, fl_d, fl_ce_n, fl_oe_n, fl_we_n */
    flctl &= ~(0x7U << 24);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
  } else {
    /* Select mode 0, set fl_a, Drive fl_a, fl_ce_n, fl_oe_n, fl_we_n, assert fl_ce_n */
    flctl = (location & 0x1fffeU) | (0x3U << 25) | (0x1U << 28);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Assert fl_oe_n */
    flctl |= (0x1U << 29);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Read fl_d */
    *value = (os_uint16) os_read_register_le32(handle, &pRegs->FLDATA);
    
    /* Deassert fl_oe_n */
    flctl &= ~(0x1U << 29);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Deassert fl_ce_n */
    flctl &= ~(0x1U << 28);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Stop driving fl_a, fl_ce_n, fl_oe_n, fl_we_n */
    flctl &= ~(0x3U << 25);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    os_dbgprint(9, ("methodFlashXrce4fx: *value=%lx\n",
		    (unsigned long) *value));
  }
  
  return MethodResult_Success;
}

MethodResult
methodFlashBbp(
  DeviceContext* pDevCtx,
  os_boolean     write,
  os_uint32      location,
  os_uint16*     value)
{
  BBPSPACE1REGS* pRegs = pDevCtx->hw.control.bbp;
  OS_BAR_HANDLE handle = pDevCtx->hw.controlHandle;
  os_uint32 flctl, flpage;
  os_uint32 val32;
  
  os_dbgprint(9, ("methodFlashBbp: write=%ld location=%lx *value=%lx\n",
		  (unsigned long) write, (unsigned long) location, (unsigned long) *value));
  
  flpage = location & 0x3fe0000U;
  os_write_register_le32(handle, &pRegs->FLPAGE, flpage);
  val32 = os_read_register_le32(handle, &pRegs->FLPAGE);
  
  if (write) {
    /* Select mode 0, set fl_a, Drive fl_a, fl_d, fl_ce_n, fl_oe_n, fl_we_n, assert fl_ce_n */
    flctl = (location & 0x1fffeU) | (0x7U << 24) | (0x1U << 28);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Set fl_d */
    os_write_register_le32(handle, &pRegs->FLDATA, *value);
    val32 = os_read_register_le32(handle, &pRegs->FLDATA);
    
    /* Assert fl_we_n */
    flctl |= (0x1U << 30);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Deassert fl_we_n */
    flctl &= ~(0x1U << 30);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Deassert fl_ce_n */
    flctl &= ~(0x1U << 28);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Stop driving fl_a, fl_d, fl_ce_n, fl_oe_n, fl_we_n */
    flctl &= ~(0x7U << 24);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
  } else {
    /* Select mode 0, set fl_a, Drive fl_a, fl_ce_n, fl_oe_n, fl_we_n, assert fl_ce_n */
    flctl = (location & 0x1fffeU) | (0x3U << 25) | (0x1U << 28);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Assert fl_oe_n */
    flctl |= (0x1U << 29);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Read fl_d */
    *value = (os_uint16) os_read_register_le32(handle, &pRegs->FLDATA);
    
    /* Deassert fl_oe_n */
    flctl &= ~(0x1U << 29);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Deassert fl_ce_n */
    flctl &= ~(0x1U << 28);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    /* Stop driving fl_a, fl_ce_n, fl_oe_n, fl_we_n */
    flctl &= ~(0x3U << 25);
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);
    val32 = os_read_register_le32(handle, &pRegs->FLCTL);
    
    os_dbgprint(9, ("methodFlashBbp: *value=%lx\n",
		    (unsigned long) *value));
  }
  
  return MethodResult_Success;
}

/*
** -----------------------------------------------------------------------
** Flash Reset Methods
** -----------------------------------------------------------------------
*/

MethodResult
methodFlashResetPlxbased(
  DeviceContext* pDevCtx)
{
  os_dbgprint(9, ("methodFlashResetPlxbased:\n"));

  return MethodResult_Success;
}

MethodResult
methodFlashResetXpl(
  DeviceContext* pDevCtx)
{
  XPLSPACE1REGS* pRegs = pDevCtx->hw.control.xpl;
  OS_BAR_HANDLE handle = pDevCtx->hw.controlHandle;
  os_uint32 fldata;
  
  os_dbgprint(9, ("methodFlashResetXpl:\n"));
  
  /* Drive fl_rst_n, assert fl_rst_n */
  fldata = (0x3U << 30);
  os_write_register_le32(handle, &pRegs->FLDATA, fldata);
  os_read_register_le32(handle, &pRegs->FLDATA);
  
  /* Deassert fl_rst_n */
  fldata = (0x1U << 30);
  os_write_register_le32(handle, &pRegs->FLDATA, fldata);
  os_read_register_le32(handle, &pRegs->FLDATA);

  /* Stop driving fl_rst_n */
  fldata = (0x0U << 30);
  os_write_register_le32(handle, &pRegs->FLDATA, fldata);
  os_read_register_le32(handle, &pRegs->FLDATA);

  /* Wait 20us or so for recovery from reset */
  os_delay_busy_us(20);
  
  return MethodResult_Success;
}

MethodResult
methodFlashResetXp(
  DeviceContext* pDevCtx)
{
  XPSPACE1REGS* pRegs = pDevCtx->hw.control.xp;
  OS_BAR_HANDLE handle = pDevCtx->hw.controlHandle;
  os_uint32 fldata;
  
  os_dbgprint(9, ("methodFlashResetXp:\n"));
  
  fldata = pDevCtx->shadow.regs.xp.fldata;
  
  /* Drive fl_rst_n, assert fl_rst_n */
  fldata |= (0x3U << 30);
  os_write_register_le32(handle, &pRegs->FLDATA, fldata);
  os_read_register_le32(handle, &pRegs->FLDATA);
  
  /* Deassert fl_rst_n */
  fldata &= ~(0x1U << 31);
  os_write_register_le32(handle, &pRegs->FLDATA, fldata);
  os_read_register_le32(handle, &pRegs->FLDATA);
  
  /* Stop driving fl_rst_n */
  fldata &= ~(0x1U << 30);
  os_write_register_le32(handle, &pRegs->FLDATA, fldata);
  os_read_register_le32(handle, &pRegs->FLDATA);
  
  /* Wait 20us or so for recovery from reset */
  os_delay_busy_us(20);
  
  return MethodResult_Success;
}

MethodResult
methodFlashResetXrc4ls(
  DeviceContext* pDevCtx)
{
  CONTROLXRC4LS* pRegs = pDevCtx->hw.control.xrc4ls;
  OS_BAR_HANDLE handle = pDevCtx->hw.controlHandle;
  os_uint32 flctl;

  os_dbgprint(9, ("methodFlashResetXrc4ls:\n"));

  flctl = pDevCtx->shadow.regs.xrc4ls.flctl;

  /* Drive fl_rst_n, assert fl_rst_n */
  flctl |= (0x1U << 27) | (0x1U << 31);
  os_write_register_le32(handle, &pRegs->FLDATA, flctl);
  os_read_register_le32(handle, &pRegs->FLDATA);

  /* Deassert fl_rst_n */
  flctl &= ~(0x1U << 31);
  os_write_register_le32(handle, &pRegs->FLDATA, flctl);
  os_read_register_le32(handle, &pRegs->FLDATA);

  /* Stop driving fl_rst_n */
  flctl &= ~(0x1U << 27);
  os_write_register_le32(handle, &pRegs->FLDATA, flctl);
  os_read_register_le32(handle, &pRegs->FLDATA);
  
  /* Wait 20us or so for recovery from reset */
  os_delay_busy_us(20);
  
  return MethodResult_Success;
}

MethodResult
methodFlashResetXrc4fx(
  DeviceContext* pDevCtx)
{
  XRC4FXSPACE1REGS* pRegs = pDevCtx->hw.control.xrc4fx;
  OS_BAR_HANDLE handle = pDevCtx->hw.controlHandle;
  os_uint32 flctl;

  os_dbgprint(9, ("methodFlashResetXrc4fx\n"));
  
  flctl = pDevCtx->shadow.regs.xrc4fx.flctl;
  
  /* Drive fl_rst_n, assert fl_rst_n */
  flctl |= (0x1U << 27) | (0x1U << 31);
  os_write_register_le32(handle, &pRegs->FLCTL, flctl);
  os_read_register_le32(handle, &pRegs->FLCTL);
  
  /* Deassert fl_rst_n */
  flctl &= ~(0x1U << 31);
  os_write_register_le32(handle, &pRegs->FLCTL, flctl);
  os_read_register_le32(handle, &pRegs->FLCTL);
  
  /* Stop driving fl_rst_n */
  flctl &= ~(0x1U << 27);
  os_write_register_le32(handle, &pRegs->FLCTL, flctl);
  os_read_register_le32(handle, &pRegs->FLCTL);
  
  os_delay_busy_us(20);
  
  return MethodResult_Success;
}

MethodResult
methodFlashResetXrce4fx(
  DeviceContext* pDevCtx)
{
  XRCE4FXSPACE1REGS* pRegs = pDevCtx->hw.control.xrce4fx;
  OS_BAR_HANDLE handle = pDevCtx->hw.controlHandle;
  os_uint32 flctl;

  os_dbgprint(9, ("methodFlashResetXrce4fx:\n"));
  
  flctl = pDevCtx->shadow.regs.xrce4fx.flctl;
  
  /* Drive fl_rst_n, assert fl_rst_n */
  flctl |= (0x1U << 27) | (0x1U << 31);
  os_write_register_le32(handle, &pRegs->FLCTL, flctl);
  os_read_register_le32(handle, &pRegs->FLCTL);
  
  /* Deassert fl_rst_n */
  flctl &= ~(0x1U << 31);
  os_write_register_le32(handle, &pRegs->FLCTL, flctl);
  os_read_register_le32(handle, &pRegs->FLCTL);
  
  /* Stop driving fl_rst_n */
  flctl &= ~(0x1U << 27);
  os_write_register_le32(handle, &pRegs->FLCTL, flctl);
  os_read_register_le32(handle, &pRegs->FLCTL);
  
  os_delay_busy_us(20);
  
  return MethodResult_Success;
}

MethodResult
methodFlashResetBbp(
  DeviceContext* pDevCtx)
{
  BBPSPACE1REGS* pRegs = pDevCtx->hw.control.bbp;
  OS_BAR_HANDLE handle = pDevCtx->hw.controlHandle;
  os_uint32 flctl;

  os_dbgprint(9, ("methodFlashResetBbp\n"));
  
  flctl = pDevCtx->shadow.regs.bbp.flctl;
  
  /* Drive fl_rst_n, assert fl_rst_n */
  flctl |= (0x1U << 27) | (0x1U << 31);
  os_write_register_le32(handle, &pRegs->FLCTL, flctl);
  os_read_register_le32(handle, &pRegs->FLCTL);
  
  /* Deassert fl_rst_n */
  flctl &= ~(0x1U << 31);
  os_write_register_le32(handle, &pRegs->FLCTL, flctl);
  os_read_register_le32(handle, &pRegs->FLCTL);
  
  /* Stop driving fl_rst_n */
  flctl &= ~(0x1U << 27);
  os_write_register_le32(handle, &pRegs->FLCTL, flctl);
  os_read_register_le32(handle, &pRegs->FLCTL);
  
  os_delay_busy_us(20);
  
  return MethodResult_Success;
}

